// Enumerate serial ports using the SetupAPI.
//
// Windows 95 and NT4 notes:
// The app may be linked to SetupAPI.lib even if Win 95 or NT4 must be supported when:
// - IE or specific service packs has been installed on the user's machine (SetupAPI.dll is present)
// - The app is build with delay loading for SetupAPI.dll and the OS version is checked
//   before calling any setup functions. No Setup functions are called when Init() didn't find the DLL. 
//
// Windows 9x notes:
// - Must use ANSI versions (Unicode versions return ERROR_CALL_NOT_IMPLEMENTED)
// - Must use port class rather than serial interface class with SetupDiGetClassDevs.
// 
// Tested with Windows 2000 SP4, XP SP2, 7, 98SE (Virtual PC 2007)
//
// SetupAPI does not work with NT4!
// The SetupAPI DLL may be present, but the functions used here can't be used to enumerate the
//  serial ports. Windows NT4 contains a registry key with the ports GUID, but this key only
//  contains the value names "Class" and "Install32" and no subkeys with device information.
// Therefore, the provided function EnumPortsNT4() must be used with NT4.
//
// Author: Jochen Arndt <jochen.arndt@t-online.de>

#include "stdafx.h"
#include "McrSerialEnumDevices.h"
#include "DebugOutput.h"

//#include <setupapi.h>								// included by EnumSerial.h
//#include <WinIoCtl.h>								// Pre-defined GUIDs

// Size for property buffer in bytes.
// NOTE: Max. string length is half size with Unicode! 
#define PROP_BUF_SIZE	1024

//#if _SETUPAPI_VER > 0x0400
//#pragma comment(lib, "setupapi")					// link setupapi of Win 95 and NT4 must be not supported
//#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

CEnumDevInfo::CEnumDevInfo()
{
	m_nStatus = 0;
	m_nPortNum = -1;
}

CEnumDevices::CEnumDevices()
{
	m_bIsNT = TRUE;									// assume Windows NT4 or later
//#if defined(_WIN32_WINNT) && _WIN32_WINNT < 0x0500	// NT4 must be supported
//	m_bIsNT4 = FALSE;								// assume not NT4
//#endif

//#if _SETUPAPI_VER < 0x0501							// late binding; get pointers to used setupapi functions
//	m_hinstSetup = NULL;							// setupapi.dll instance handle
//	m_GetCustomDeviceProperty = NULL;
//#if _SETUPAPI_VER <= 0x0400							// Windows 95 or NT4 must be supported
//	m_GetClassDevsA = NULL;
//	m_DestroyDeviceInfoList = NULL;
//	m_EnumDeviceInfo = NULL;
//	m_GetDeviceRegistryProperty = NULL;
//#endif // _SETUPAPI_VER <= 0x0400
//#endif // _SETUPAPI_VER < 0x0501
}

CEnumDevices::~CEnumDevices()
{
//#if _SETUPAPI_VER < 0x0501
//	if (NULL != m_hinstSetup)
//		::FreeLibrary(m_hinstSetup);
//#endif
}

// Get COM port file name from port number.
/*static*/ bool CEnumDevices::GetFileName(int nPortNum, TCHAR *tcStr, int iLength)
{
	int ii;
	wchar_t wcTempBuff[256];

	if (nPortNum > 0)
	{
		if (nPortNum >= 10)		
			swprintf_s(wcTempBuff,sizeof(iLength)/sizeof(TCHAR), L"\\\\.\\COM%d", nPortNum);
		else
			swprintf_s(wcTempBuff,sizeof(iLength)/sizeof(TCHAR), L"COM%d", nPortNum);

		for (ii=0; ii<256; ii++)
		{
			if (wcTempBuff[ii] != 0)
			{
				tcStr[ii] = (TCHAR)wcTempBuff[ii];
			} else
			{
				tcStr[ii] = 0;
				break;
			}
		}
	} else
	{
		tcStr[0] = 0;
	}

	return nPortNum >= 0 ? true : false;
}

// Get port number form COM port file name.
// Return value:
// -1 : Invalid name
//  0 : Empty name treated as "no port".
// >0 : Port number
/*static*/ int CEnumDevices::GetPortFromName(LPCTSTR lpszPort)
{
	#ifdef _DEBUG_ERROR
	if (lpszPort == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!GetPortFromName: null port"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	int nPortNum = -1;
	if (lpszPort != NULL)
	{
		if (*lpszPort == _T('\0'))
			nPortNum = 0;
		else
		{
			LPCTSTR s = lpszPort;
			if (_tcsncmp(s, _T("\\\\.\\"), 4) == 0)
				s += 4;
			if (_tcsnicmp(s, _T("COM"), 3) == 0)
				nPortNum = _tstoi(s + 3);
			if (nPortNum <= 0)
			{
				#ifdef _DEBUG_ERROR
				{
					TCHAR tcDebugBuff[256];
					swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!Invalid COM port file name %s\n"), lpszPort);
					cDebugOutput->DebugMsg(tcDebugBuff);
				}
				#endif

				nPortNum = -1;
			}
		}
	}
	return nPortNum;
}

BOOL CEnumDevices::InitSetupApi()
{
	BOOL bResult = TRUE;

//#if WINVER < 0x0500
//	OSVERSIONINFO VerInfo;							// Get Windows version information.
//	VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
//	::GetVersionEx(&VerInfo);
//	m_bIsNT = (VerInfo.dwPlatformId == VER_PLATFORM_WIN32_NT);
//#if defined(_WIN32_WINNT) && _WIN32_WINNT < 0x0500	// NT4 must be supported
//	m_bIsNT4 = (VerInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) && (VerInfo.dwMajorVersion == 4); 
//	if (m_bIsNT4)									// Setup API not used with NT4
//		return TRUE;
//#endif
//#endif

//#if _SETUPAPI_VER < 0x0501							// late binding; get pointers to used setupapi functions
//	if (NULL == m_hinstSetup)
//	{
//		m_hinstSetup = ::LoadLibrary(_T("setupapi"));
//		if (m_hinstSetup == NULL)
//		{
//			bResult = FALSE;
//			TRACE0("SetupApi.dll not found\n");
//		}
//		else
//		{
//
//			m_GetCustomDeviceProperty =				// requires setupapi >= 0x0501
//#ifdef _UNICODE
//				(LP_GetCustomDeviceProperty)::GetProcAddress(m_hinstSetup, "SetupDiGetCustomDevicePropertyW");
//#else
//				(LP_GetCustomDeviceProperty)::GetProcAddress(m_hinstSetup, "SetupDiGetCustomDevicePropertyA");
//#endif
////#if _SETUPAPI_VER <= 0x0400							// Windows 95 or NT4 must be supported
////			m_GetClassDevsA =
////				(LP_GetClassDevsA)::GetProcAddress(m_hinstSetup, "SetupDiGetClassDevsA");
////			m_EnumDeviceInfo =
////				(LP_EnumDeviceInfo)::GetProcAddress(m_hinstSetup, "SetupDiEnumDeviceInfo");
////			m_DestroyDeviceInfoList =
////				(LP_DestroyDeviceInfoList)::GetProcAddress(m_hinstSetup, "SetupDiDestroyDeviceInfoList");
////#ifdef _UNICODE
////			m_GetDeviceRegistryProperty = m_bIsNT ?
////				(LP_GetDeviceRegistryProperty)::GetProcAddress(m_hinstSetup, "SetupDiGetDeviceRegistryPropertyW") :
////				(LP_GetDeviceRegistryProperty)::GetProcAddress(m_hinstSetup, "SetupDiGetDeviceRegistryPropertyA");
////#else
////			m_GetDeviceRegistryProperty = 
////				(LP_GetDeviceRegistryProperty)::GetProcAddress(m_hinstSetup, "SetupDiGetDeviceRegistryPropertyA");
////#endif
////			if (m_GetClassDevsA == NULL ||
////				m_EnumDeviceInfo == NULL ||
////				m_DestroyDeviceInfoList == NULL ||
////				m_GetDeviceRegistryProperty == NULL)
////			{
////				bResult = FALSE;					// required function not found
////				TRACE("Failed to locate required setupapi.dll functions\n");
////			}
////#endif // _SETUPAPI_VER <= 0x0400
//		}
//	}
//#endif // _SETUPAPI_VER < 0x0501
	return bResult;
}

// Generate error message for SetupAPI errors.
// NOTE: Must use special conversion macro HRESULT_FROM_SETUPAPI.
void CEnumDevices::Error(LPCTSTR lpszFunction, bool bMsg /*= false*/) const
{
	#ifdef _DEBUG_ERROR
	if (lpszFunction == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::Error null lpszFunction"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	DWORD dwErr = ::GetLastError();
	if (dwErr)
	{
		LPVOID lpMsgBuf;
		if (::FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			HRESULT_FROM_SETUPAPI(dwErr),			// special conversion for SetupAPI errors
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default user language
			reinterpret_cast<LPTSTR>(&lpMsgBuf),
			0,
			NULL))
		{
			#ifdef _DEBUG_ERROR
			if (lpszFunction == NULL)
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!SetupAPI error %d (%#X) in %s: %s\n"), 
					                                                     dwErr, dwErr, lpszFunction, 
					                                                     reinterpret_cast<LPCTSTR>(lpMsgBuf));
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif

			if (bMsg && dwErr != ERROR_CALL_NOT_IMPLEMENTED)
			{
// kjf xxx yyy zzz add DebugMsg
				//TCHAR strMsg[512];
				//strMsg.Format(_T("SetupAPI error in %s:\n%s"), 
				//	lpszFunction, 
				//	reinterpret_cast<LPCTSTR>(lpMsgBuf));
				//AfxMessageBox(static_cast<LPCTSTR>(strMsg));
			}
			::LocalFree(lpMsgBuf);
		}
	}
}

// With Windows 2K/XP and later we can use GUID_DEVINTERFACE_COMPORT with DIGCF_DEVICEINTERFACE.
// This does not work with Windows 9x where we must use
//  GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR which returns also LPT ports.
// We can always use the ANSI version here, because the PnP enum/name parameter isn't used.
HDEVINFO CEnumDevices::GetClassDevs(const GUID* lpGUID, BOOL bPresent /*= TRUE*/)
{
	DWORD dwFlags = bPresent ? DIGCF_PRESENT : 0;	// only present devices or all that has been seen
	
	//if (lpGUID == &GUID_DEVINTERFACE_COMPORT)		// Fixed GUID for COM ports (NT only) 
	//	dwFlags |= DIGCF_DEVICEINTERFACE;
	//else if (lpGUID == NULL)						// No specific GUID (all classes)
		dwFlags |= DIGCF_ALLCLASSES;
//#if _SETUPAPI_VER <= 0x0400							// Windows 95 or NT4
//	HDEVINFO hDevInfo = INVALID_HANDLE_VALUE;
//	if (m_GetClassDevsA == NULL)
//		::SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
//	else
//		hDevInfo = m_GetClassDevsA(lpGUID, NULL, NULL, dwFlags); 
//#else
	HDEVINFO hDevInfo = ::SetupDiGetClassDevsA(lpGUID, NULL, NULL, dwFlags); 
//#endif // _SETUPAPI_VER <= 0x0400
	if (hDevInfo == INVALID_HANDLE_VALUE)
		Error(_T("SetupDiGetClassDevs"));
	return hDevInfo;
}

BOOL CEnumDevices::GetPropertyString(HDEVINFO hDevInfo, PSP_DEVINFO_DATA pDevInfoData, DWORD dwProperty, TCHAR *pstrProperty, int iNumberOfElements)
{
	TCHAR *ptcProperty;

	#ifdef _DEBUG_ERROR
	if (hDevInfo == INVALID_HANDLE_VALUE)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetPropertyString invalid handle hDevInfo"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_ERROR
	if (pDevInfoData == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetPropertyString null pDevInfoData"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_ERROR
	if (dwProperty >= SPDRP_MAXIMUM_PROPERTY)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetPropertyString invalid dwProperty: %d"), dwProperty);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	BOOL bResult = FALSE;
	DWORD dwDataType = 0;							// registry data type (e.g. REG_SZ)
	DWORD dwReqSize = 0;							// retrieve required buffer size
	BYTE lpszBuf[PROP_BUF_SIZE];
	
//#if _SETUPAPI_VER <= 0x0400							// Windows 95 or NT4
//	if (m_GetDeviceRegistryProperty == NULL)
//		::SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
//	else
//	{
//		bResult = m_GetDeviceRegistryProperty(
//			hDevInfo,								// <- device info set handler
//			pDevInfoData,							// <- enumerated device info data
//			dwProperty,								// <- Description and name as displayed in the hardware config
//			&dwDataType,							// -> registry data type (e.g. REG_SZ), optional
//			lpszBuf,								// -> output buffer filled with registry data
//			PROP_BUF_SIZE,							// <- output buffer size
//			&dwReqSize);							// -> required size, optional
//		if (bResult && dwDataType == REG_SZ)		// just to be sure ...
//		{
//#ifdef _UNICODE
//			if (!m_bIsNT)							// ANSI version has been called
//				str = reinterpret_cast<LPCSTR>(lpszBuf);
//			else									// Unicode version has been called
//				str = reinterpret_cast<LPCTSTR>(lpszBuf);
//#else
//			str = reinterpret_cast<LPCSTR>(lpszBuf);
//#endif
//		}
//	}
//#else
#ifdef _UNICODE
	if (m_bIsNT)									// call the Unicode version
	{
		bResult = ::SetupDiGetDeviceRegistryProperty(
			hDevInfo,								// <- device info set handler
			pDevInfoData,							// <- enumerated device info data
			dwProperty,								// <- Description and name as displayed in the hardware config
			&dwDataType,							// -> registry data type (e.g. REG_SZ), optional
			lpszBuf,								// -> output buffer filled with registry data
			PROP_BUF_SIZE,							// <- output buffer size
			&dwReqSize);							// -> required size, optional
		if (bResult && dwDataType == REG_SZ)		// just to be sure ...
		{
			ptcProperty = reinterpret_cast<TCHAR *>(lpszBuf);
			wcscpy_s(pstrProperty, iNumberOfElements, ptcProperty);
		}
	}
	else											// call the ANSI version with Windows 9x
#endif
	{
		bResult = ::SetupDiGetDeviceRegistryPropertyA(
			hDevInfo,								// <- device info set handler
			pDevInfoData,							// <- enumerated device info data
			dwProperty,								// <- Description and name as displayed in the hardware config
			&dwDataType,							// -> registry data type (e.g. REG_SZ), optional
			lpszBuf,								// -> output buffer filled with registry data
			PROP_BUF_SIZE,							// <- output buffer size
			&dwReqSize);							// -> required size, optional
		if (bResult && dwDataType == REG_SZ)		// just to be sure ...
		{
			int ii;

			ptcProperty = reinterpret_cast<TCHAR *>(lpszBuf);
			for (ii=0; ii<iNumberOfElements; ii++)
				pstrProperty[ii] = ptcProperty[ii];
			//wcscpy_s(pstrProperty, iNumberOfElements, ptcProperty);
		}

	}

//#endif // _SETUPAPI_VER <= 0x0400

	if (!bResult)
	{
// kjf xxx yyy zzz
		// add DebugMsg
		//CString s;
		//s.Format(_T("SetupDiGetDeviceRegistryProperty(%d)"), dwProperty);
		//Error(s);
	}
	else
	{
		#ifdef _DEBUG_ERROR
		if (dwDataType != REG_SZ)
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetPropertyString dwDataType invalid size"));
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		#ifdef _DEBUG_DEVICES_INFO
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("Property %#X = %s\n"), dwProperty, pstrProperty);
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif
	}
	return bResult;
}

// Get value from 'Device Parameters' key.
// Requires Windows XP or later (API version >= 5.01).
BOOL CEnumDevices::GetCustomPropertyString(HDEVINFO hDevInfo, PSP_DEVINFO_DATA pDevInfoData, LPCTSTR lpszEntry, TCHAR *pstrProperty, int iNumberOfElements)
{
	TCHAR *ptcProperty;

	#ifdef _DEBUG_ERROR
	if (hDevInfo == INVALID_HANDLE_VALUE)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetCustomPropertyString hDevInfo invalid handle"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_ERROR
	if (pDevInfoData == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetCustomPropertyString pDevInfoData invalid handle"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_ERROR
	if ((lpszEntry == NULL) || (*lpszEntry == _T('\0')))
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetCustomPropertyString lpszEntry invalid"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	DWORD dwDataType = 0;							// registry data type (e.g. REG_SZ)
	DWORD dwReqSize = 0;							// retrieve required buffer size
	BYTE lpszBuf[PROP_BUF_SIZE];
	
//#if _SETUPAPI_VER < 0x0501							// < Windows XP
//	BOOL bResult = FALSE;
//	if (m_GetCustomDeviceProperty == NULL)
//	{
//		::SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
//	}
//	else
//	{
//		bResult = m_GetCustomDeviceProperty(		// get value from 'Device Parameters' key
//			hDevInfo, pDevInfoData,
//			lpszEntry,								// get value from entry
//			0, &dwDataType,
//			lpszBuf, PROP_BUF_SIZE, &dwReqSize);
//	}
//#else												// Windows XP or later
	BOOL bResult = ::SetupDiGetCustomDeviceProperty(
		hDevInfo, pDevInfoData,
		lpszEntry, 0, &dwDataType,
		lpszBuf, PROP_BUF_SIZE, &dwReqSize);
//#endif
	if (!bResult)
	{
// kjf xxx yyy zzz add DebugMsg
		//CString s;
		//s.Format(_T("SetupDiGetCustomDeviceProperty(%s)"), lpszEntry);
		//Error(s);
	}
	else
	{
		#ifdef _DEBUG_ERROR
		if (dwDataType != REG_SZ)
		{
			TCHAR tcDebugBuff[256];
			swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetCustomPropertyString dwDataType invalid size"));
			cDebugOutput->DebugMsg(tcDebugBuff);
		}
		#endif

		if (dwDataType == REG_SZ)
		{
			int ii;

			ptcProperty = reinterpret_cast<TCHAR *>(lpszBuf);
			for (ii=0; ii<iNumberOfElements; ii++)
				pstrProperty[ii] = ptcProperty[ii];
			//wcscpy_s(pstrProperty, iNumberOfElements, ptcProperty);

			#ifdef _DEBUG_INFO
			{
				TCHAR tcDebugBuff[256];
				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("CEnumDevices::GetCustomPropertyString Custom property %s = %s"), lpszEntry, pstrProperty);
				cDebugOutput->DebugMsg(tcDebugBuff);
			}
			#endif
		}
	}
	return bResult;
}

// Enumerate serial ports and pass data to ListBox/ComboBox using call back function.
BOOL CEnumDevices::EnumSerialPorts(EnumCallBack pCallBack, BOOL bPresent /*= TRUE*/, BOOL bPhysical /*=FALSE*/)
{
//#if defined(_WIN32_WINNT) && _WIN32_WINNT < 0x0500	// NT4 must be supported
//	if (m_bIsNT4)									// must use special function with NT4
//		return EnumSerialPortsNT4(pCallBack);
//#endif
	return EnumDevicesList(
		bPresent ? INFO_PORTS : INFO_PORTS | INFO_NON_PRESENT,
		bPhysical,
		pCallBack,
		NULL
//#if defined(_WIN32_WINDOWS) && _WIN32_WINDOWS < 0x0500 // Windows 9x must be supported
//		!m_bIsNT ?									// GUID_DEVINTERFACE_COMPORT not supported with Win 9x
//		&GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR :
//#endif
//		&GUID_DEVINTERFACE_COMPORT
        );
}

// Enumerate devices and pass data using call back function.
BOOL CEnumDevices::EnumDevicesList(unsigned nInfo, BOOL bPhysical, EnumCallBack pCallBack, const GUID* lpGUID)
{
	BOOL bIsSuccess;

	#ifdef _DEBUG_ERROR
	if (pCallBack == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::EnumDevices pCallBack null"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	#ifdef _DEBUG_ERROR
	if (pCallBack == NULL)
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("!!!CEnumDevices::GetCustomPropertyString pCallBack invalid handle"));
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	DWORD nNumDev = 0;								// count interfaces
	HDEVINFO hDevInfo = INVALID_HANDLE_VALUE;		// device infoset handle
	SP_DEVINFO_DATA	DevInfoData;					// device info data for enumerated device

	::ZeroMemory(&DevInfoData, sizeof(SP_DEVINFO_DATA));
	DevInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

	#ifdef _DEBUG_INFO
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("EnumDevices(%#X)"), nInfo);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	BOOL bResult = InitSetupApi();
	if (bResult)
	{
		hDevInfo = GetClassDevs(lpGUID, !(nInfo & INFO_NON_PRESENT));
		if (hDevInfo == INVALID_HANDLE_VALUE)
			bResult = FALSE;
	}

	//try
	{
		while (bResult)								// enumerate devices
		{
//#if _SETUPAPI_VER <= 0x0400							// Windows 95 or NT4
//			bResult = m_EnumDeviceInfo(				// get device info
//				hDevInfo, nNumDev++, &DevInfoData);
//#else
			bResult = ::SetupDiEnumDeviceInfo(		// get device info
				hDevInfo, nNumDev++, &DevInfoData);
//#endif
			if (!bResult)
			{
				if (::GetLastError() == ERROR_NO_MORE_ITEMS)
					bResult = TRUE;					// no more devices
				else
					Error(_T("SetupDiEnumDeviceInfo()"));
				break;
			}

			CEnumDevInfo *pInfo = new CEnumDevInfo;


			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_FRIENDLYNAME, pInfo->m_strName, sizeof(pInfo->m_strName)/sizeof(TCHAR));
			if ((nInfo & INFO_NAME) && bIsSuccess)
				pInfo->m_nStatus |= INFO_NAME;


			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_SERVICE, pInfo->m_strService, sizeof(pInfo->m_strService)/sizeof(TCHAR));
			if ((nInfo & INFO_SERVICE) && bIsSuccess)
				pInfo->m_nStatus |= INFO_SERVICE;

			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_MFG, pInfo->m_strManufacturer, sizeof(pInfo->m_strManufacturer)/sizeof(TCHAR));
			if ((nInfo & INFO_MFG) && bIsSuccess)
				pInfo->m_nStatus |= INFO_MFG;

			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_DEVICEDESC, pInfo->m_strDescription, sizeof(pInfo->m_strDescription)/sizeof(TCHAR));
			if ((nInfo & INFO_DESCR) && bIsSuccess)
				pInfo->m_nStatus |= INFO_DESCR;

			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_CLASS, pInfo->m_strClassName, sizeof(pInfo->m_strClassName)/sizeof(TCHAR));
			if ((nInfo & INFO_CLASS_NAME) && bIsSuccess)
				pInfo->m_nStatus |= INFO_CLASS_NAME;

			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_CLASSGUID, pInfo->m_strClassName, sizeof(pInfo->m_strClassName)/sizeof(TCHAR));
			if ((nInfo & INFO_CLASS_GUID) && bIsSuccess)
				pInfo->m_nStatus |= INFO_CLASS_GUID;

			bIsSuccess = GetPropertyString(hDevInfo, &DevInfoData, SPDRP_ENUMERATOR_NAME, pInfo->m_strPnPName, sizeof(pInfo->m_strPnPName)/sizeof(TCHAR));
			if ((nInfo & INFO_PNP_NAME) && bIsSuccess)
				pInfo->m_nStatus |= INFO_PNP_NAME;

			bIsSuccess = GetCustomPropertyString(hDevInfo, &DevInfoData, _T("PortName"), pInfo->m_strPort, sizeof(pInfo->m_strPort)/sizeof(TCHAR));
			if ((nInfo & INFO_PORT_NAME) && bIsSuccess)
				pInfo->m_nStatus |= INFO_PORT_NAME;

			// Serial port specific handling:
			//  Get the port name 'COMx' and the user friendly name.
			//  The used function for the port name requires a API version >= 5.01 (Windows XP).
			//  With older versions, we will extract the port name from the user friendly name.
			//  Extraction assumes that the port name is provided in parentheses as '(COMx)'.
			//  Looking for 'COM' only does not always work (e.g. with German Win 98 where
			//  the user friendly name is 'COM Anschlu (COM1)')!

			if ((pInfo->m_nStatus & INFO_NAME) &&
				!(pInfo->m_nStatus & INFO_PORT_NAME))
			{
				const wchar_t *tcStartPos = wcsstr(pInfo->m_strName, L"(COM");
				if (tcStartPos)
				{
					const wchar_t *tcEndPos = wcsstr(pInfo->m_strName, L")");

					wchar_t *tcDestPos = pInfo->m_strPort;
					if (tcEndPos)
					{
						int ii, jj;
						TCHAR tcPortNumber[16];

						// we add one to get past the '('
						tcStartPos++;

						jj = ((int)(tcStartPos - pInfo->m_strName))/sizeof(TCHAR);

						memset(pInfo->m_strPort, 0, sizeof(pInfo->m_strPort));
						for (ii=0; ii<(sizeof(pInfo->m_strPort)/sizeof(TCHAR)); ii++)
						{
							*tcDestPos++ = *tcStartPos++;

							if (tcStartPos >= tcEndPos)
							{
								break;
							}
						}

						// need the port number
						tcDestPos = pInfo->m_strPort;

						memset(tcPortNumber, 0, sizeof(tcPortNumber));
						for (ii=0, jj=0; ii<(sizeof(pInfo->m_strPort)/sizeof(TCHAR)); ii++)
						{
							if ((*tcDestPos >= '0') && (*tcDestPos <= '9'))
							{
								tcPortNumber[jj++] = *tcDestPos;
							}

							tcDestPos++;

							if (!*tcDestPos)
								break; // got to the end of the string
						}

						pInfo->m_nPortNum = _tstoi(tcPortNumber);

						pInfo->m_nStatus |= INFO_PORT_NAME;
					}
				}
			}

//			if ((pInfo->m_nStatus & INFO_PORT_NAME) &&	// found a serial port
//				pInfo->m_strPort.Left(3) == _T("COM"))
//				&& _istdigit(strPort.GetAt(4)))
			if (pInfo->m_nStatus & INFO_PORT_NAME)
			{
				pInfo->m_nPortNum = _tstoi(pInfo->m_strPort + 3);

				// Getting the service type fails with Windows 9x.
				// If the service type is not "Serial", it is probably a virtual
				//  COM port (e.g. USB to serial converter or internal modem card).
				// This is indicated by adding a special bit to the port number.
				if (pInfo->m_nStatus & INFO_SERVICE)
				{
					TCHAR tcStrService[512];
					int ii;
					for (ii=0; ii<512; ii++)
					{
						if (pInfo->m_strService[ii])
						{
							tcStrService[ii] = tolower(pInfo->m_strService[ii]);
						} else
						{
							break;
						}
					}

					if (pInfo->m_nPortNum && (wcscmp(pInfo->m_strService,_T("Serial"))))
						pInfo->m_nPortNum |= DATA_VIRTUAL_MASK;
				}
				else
					wcscpy_s(pInfo->m_strService, sizeof(pInfo->m_strService)/sizeof(TCHAR),_T("Serial"));
			}
			pCallBack(pInfo, bPhysical);

			delete pInfo;

		} // while (bResult)
	}
	//catch (CMemoryException *e)
	//{
	//	e->Delete();
	//	bResult = FALSE;
	//}
	if (hDevInfo != INVALID_HANDLE_VALUE)
	{
//#if _SETUPAPI_VER <= 0x0400
//		m_DestroyDeviceInfoList(hDevInfo);
//#else
		::SetupDiDestroyDeviceInfoList(hDevInfo);
//#endif
	}

	#ifdef _DEBUG_INFO
	{
		TCHAR tcDebugBuff[256];
		swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("CEnumDevices::EnumDevices %d devices found\n"), --nNumDev);
		cDebugOutput->DebugMsg(tcDebugBuff);
	}
	#endif

	return bResult;
}

//#if defined(_WIN32_WINNT) && _WIN32_WINNT < 0x0500
//// Enumerate serial ports by searching in the registry key "HKLM\Hardware\DeviceMap\SerialComm".
//// Check for NT4 before calling this function!
//BOOL CEnumDevices::EnumSerialPortsNT4(EnumCallBack pCallBack)
//{
//	ASSERT(pCallBack != NULL);
//
//	int nPorts = 0;
//	HKEY hKey;
//	TCHAR lpszDevName[40], lpszPortName[20];
//	DWORD dwType;
//
//	TRACE0("EnumSerialPortsNT4\n");
//	LONG lResult = ::RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
//		_T("Hardware\\DeviceMap\\SerialComm"), 
//		0, KEY_READ, &hKey);
//	if (lResult != ERROR_SUCCESS) 
//	{
//		TRACE1(" Failed. RegOpenKeyEx returns %#X\n", lResult);
//	}
//	while (lResult == ERROR_SUCCESS) 
//	{
//		DWORD dwDevNameSize = sizeof(lpszDevName);
//		DWORD dwPortNameSize = sizeof(lpszPortName);
//		lResult = ::RegEnumValue(hKey, nPorts, 
//			lpszDevName, &dwDevNameSize,			// value name (Serial<n>)
//			NULL, &dwType, 
//			reinterpret_cast<LPBYTE>(lpszPortName),	// value data (COMx)
//			&dwPortNameSize);
//		if (lResult != ERROR_SUCCESS)
//		{
//			VERIFY(::RegCloseKey(hKey) == ERROR_SUCCESS);
//			if (lResult != ERROR_NO_MORE_ITEMS)
//				TRACE1(" Failed. RegEnumValue returns %#X\n", lResult);
//		}
//		else
//		{
//			++nPorts;
//			ASSERT(dwType == REG_SZ);
//
//			#ifdef _DEBUG_INFO
//			{
//				TCHAR tcDebugBuff[256];
//				swprintf_s(tcDebugBuff,sizeof(tcDebugBuff)/sizeof(TCHAR),_T("CEnumDevices::EnumSerialPortsNT4 %s, device %s\n"), lpszPortName, lpszDevName);
//				cDebugOutput->DebugMsg(tcDebugBuff);
//			}
//			#endif
//
//			CEnumDevInfo Info;
//			Info.m_nStatus = INFO_NAME | INFO_PORT_NAME;
//			Info.m_nPortNum = _tstoi(lpszPortName+3);
//			Info.m_strPort = Info.m_strName = lpszPortName;
//			Info.m_strService = _T("Serial");
//			pCallBack(&Info);
//		}
//	}
//	TRACE1(" %d serial ports found\n", nPorts);
//	return (lResult == ERROR_SUCCESS || lResult == ERROR_NO_MORE_ITEMS);
//}
//#endif

//#if 0
//// Enumerate serial ports using GetDefaultCommConfig.
//// Just for reference.
//void CEnumDevices::EnumSerialPortsOld(EnumCallBack pCallBack, int nMaxPort /*= 10*/)
//{
//	ASSERT(pCallBack != NULL);
//		
//	int nPorts = 0;
//	COMMCONFIG CC;
//	TCHAR strPortName[512];
//	for (int i = 1 ; i <= nMaxPort; i++)
//	{
//		bool bPortExist = false;
//		strPortName.Format(_T("COM%d"), i);			// name of port
//		// GetDefaultCommConfig() accepts only COM1 to COM9 without extended syntax!
//		// Requires about 10 ms per port.
//		// Works also for opened ports.
//		if (i < 10)
//		{
//			DWORD dwSize = sizeof(COMMCONFIG);		// must be re-initialized inside loop
//			if (::GetDefaultCommConfig(static_cast<LPCTSTR>(strPortName), &CC, &dwSize))
//			{
//				if (CC.dwProviderSubType == PST_RS232)
//					bPortExist = true;				// add port to list
//			}
//		}
//		else										// try to open the port
//		{
//			TCHAR strPortFile[512];
//			strPortFile.Format(_T("\\\\.\\COM%d"), i);	// file name of port
//			HANDLE hPort = ::CreateFile(
//				strPortFile, 
//				GENERIC_READ | GENERIC_WRITE, 
//				0, NULL, OPEN_EXISTING, 0, NULL);
//			if (hPort != INVALID_HANDLE_VALUE)
//			{
//				bPortExist = true;
//				VERIFY(::CloseHandle(hPort));
//			}
//		}
//		if (bPortExist)
//		{
//			CEnumDevInfo Info;
//			Info.m_nStatus = INFO_NAME | INFO_PORT_NAME;
//			Info.m_nPortNum = i;
//			Info.m_strPort = Info.m_strName = strPortName;
//			Info.m_strService = _T("Serial");
//			pCallBack(&Info);
//			++nPorts;
//		}
//	}
//}
//#endif
