#pragma once




// the protocol is very simplistic using only ASCII printable characters.
//
// A command with no data is just the letter.
// A command with data is the letter followed by <data>.
//
// A response with no data is just the letter.
// A response with data is the letter followed by <data>.
//
// Data within the packet is variable length and comma delimited.
// There are no leading zeros. All integer values are in decimal.
//
// For example, the current date/time is :
// Feb 23, 2012 15:35:59.5000 the fractional second is in microseconds
// PC sends 'a' to the AVR
// The AVR responds (ASCII characters): b<2,23,2012,15,35,39,5000>
//
// Most of the time, this is a half duplex protocol, however,
// full duplex is allowed.


// In addition to the commands below, the AVR can toggle the DSR modem control line.
// When the PC detects the toggling of the modem control line it latches the
// current value of the high performance counter and performs top of second functions.
//

#define COMMAND_NOT_RECOGNIZED 1			// 1 ==command not recognized
#define BAD_PARAMETER          2            // 2 ==bad parameter
#define TOO_MANY_PARAMETERS    3			// 3 ==too many parameters
#define NOT_ENOUGH_PARAMETERS  4			// 4 ==not enough parameters

#define DATE_TIME_REQUEST		    'a'	// Read the date/time, initiated by AVR or PC
#define DATE_TIME_RESPONSE		    'b'	//
								    	// Request: "a"
								    	// Response: "b<month,day,year,hour,minute,second,microseconds>"
#define PC_SET_DATE_TIME            'c' // Set the date/time in the AVR.
                                        // Request: "c<month,day,year,hour,minute,second,microseconds>"
										// Response: "Y" (ie. ACK)
#define PC_STATUS_REQUEST       	'd'	// Read the AVR Status, initiated by the PC
#define PC_STATUS_RESPONSE       	'e'	//
								    	// Request: "d"
                                        // Response "e< same as AVR_NOTIFY_ALARM request >
#define AVR_NOTIFY_ALARM        	'f'	// Alarm Status, initiated by AVR, one of the values below changed
										// Request: "f<GeneralLockStatus,RtcStatus,GpsStatus,tcStatus,hsoStatus,hsoAchieve,0,0,0,0,hwStatus,verMajor,verMinor,verRevision,0,0,0,0>"
                                        // GeneralLockStatus ==0, no valid reference
										//                   ==1, lock gain alarm (from no lock to either time code lock or GPS lock)
										//                   ==2, lock loss alarm (from GPS lock or time code lock to no lock on either time code or GPS)
										//
										// Lock Status is:
                                        //                  LOCK_INVALID         0x00   // never had a lock
                                        //                  LOCK_NOT_PRESENT     0x01   // had a lock but lost it
                                        //                  LOCK_LOST            0x02   // notify, one time, that lock was lost
                                        //                  LOCK_RESTORED        0x03   // notify, one time, that lock was restored
                                        //                  LOCK_PRESENT         0x04   // currently locked to a reference
                                        //
                                        // RtcStatus         == Lock Status
   										// GpsStatus         == Lock Status
										// tcStatus          == Lock Status
										// hsoStatus         == Lock Status
                                        // hsoAchieve        == HSO achievement Level, 0-16
                                        // Reserved1,2,3,4   == 0
                                        // hwStatus          == 0, tbd
                                        // verMajor          == firmware version major
                                        // verMinor          == firmware version minor
                                        // verRevision       == firmware version revision
                                        // Reserved1,2,3,4   == 0
										// Response: "Y" (ie. ACK)
#define AVR_FRAME_ALARM         	'g'	// Frame Alarm, initiated by AVR, start/end of a time code frame.
										// Response: "Y" (ie. ACK)
#define AVR_NMEA_PACKET 			'h'	// NMEA packet, initiated by AVR
										// Request: "h<nmea>"
										// where "nmea" is the nmea packet as received from the GPS receiver
										// with the exception of ZDA which is auto-generated by the AVR
										// at top of second.
										// Response: "Y" (ie. ACK)
#define AVR_NOTIFY_TOP_OF_SECOND	'i'	// NOTE::: The AVR can EITHER toggle the DSR modem control line OR
										// send 'i', NOT BOTH.
										// Latch high performance counter, initiated by AVR
                                        // PC reads and saves the current value of the high performance counter
                                        // This causes the same function as the AVR toggling the DSR modem control line.
                                        // The PC will then read various information related to top of second,
                                        // such as current status, time code frame from the previous second,
                                        // current time, etc.
                                        // Request: "h"
										// Response: "Y" (ie. ACK)
#define PC_READ_TIME_CODE_REQUEST	'j'	// read the last time code frame decoded, initiated by PC
#define PC_READ_TIME_CODE_RESPONSE	'k'	//
										// Request: "j"
										// Response: "k<tcStatus,hwStatus,a,b,...>"
                                        // Where a,b,... value is ascii representation of the hex time code data
                                        //               (always 13 values, zero filled at end if time code frame has less than 13 bytes).
                                        // for example "255" is the hex byte 0xFF
#define PC_RECALIBRATE_REQUEST  	'l'	// Recalibrate time code input, initiated by the PC
										// Request: "l"
										// Response: "Y" (ie. ACK)
#define PC_SET_CONFIG               'm' // Set misc. configuration parameters
                                        // request: "m<ucIrigTimeCodeDateAvailable,ucSmpteTimeCodeDateAvailable,bIsTzBiasPositive,tzBiasSeconds,dstBiasSeconds,bDstIsUtcReferenced,
                                        //             dstStandardMonth,dstStandardDate,dstStandardYear,dstStandardDayOfWeek,
                                        //             dstStandardHour,dstStandardMinute,dstStandardSecond,
										//             dstDaylightMonth,dstDaylightDate,dstDaylightYear,dstDaylightDayOfWeek
                                        //             dstDaylightHour,dstDaylightMinute,dstDaylightSecond>"
                                        // where:
                                        // ucIrigTimeCodeDateAvailable  == 0, IRIG  time code does not have date information available
                                        //                              == 1, IRIG  time code has date using IEEE 1344 standard
										// ucSmpteTimeCodeDateAvailable == 0, SMPTE time code does not have date information available
										//                              == 1, SMPTE time code has date using Leitch standard
										//                              == 2, SMPTE time code has date using 309M with time zone
										//                              == 3, SMPTE time code has date using 309M no time zone
                                        // bIsTzBiasPositive            == 0, time zone bias is negative
                                        //                              == 1, time zone bias is positive
                                        // tzBiasSeconds                == time zone offset in seconds, up to one day
                                        // dstBiasSeconds               == daylight saving bias in seconds, up to 65535
                                        // bDstIsUtcReferenced          == 0, DST occurs based on local time
                                        //                              == 1, DST occurs based on UTC time
                                        //But wDayOfWeek and wDay take some explanation:
                                        //When the stored dates are absolute (wYear == 1) wDayOfWeek is always 0 and wDay is the day of the month.
                                        //For instance, November 15th at 07:10:30is stored as 1-11-0-15-7-10-30 (leaving off the unused milliseconds field).
                                        //
                                        //When relative (wYear == 0) wDayOfWeek is a # corresponding to the day of the week,
                                        //where Sunday = 0, Monday = 1, ... Saturday = 6
                                        //and wDay is the ordinal # for the occurrence of that day in the month.
                                        //
                                        //So, the 1st Sunday in November, 2 AM is stored as 0-11-0-1-2-0-0-0
                                        //That's wYear-wMonth-wDayOfWeek-wDay-wHour-wMinute-wSecond
                                        //         0     11       0        1    2      0       0
                                        //where 0 in wDayOfWeek means Sunday and 1 in wDay means the 1st one in the month.
                                        //
                                        //and the 4th Monday in October, 06:07:08 is
                                        //         0     10       1        4    6      7       8
                                        //where 1 in wDayOfWeek means Monday and 4 in wDay means the 4th one in the month.
                                        //
										// dstStandardMonth
										// dstStandardDate
										// dstStandardYear
										// dstStandardDayOfWeek
                                        // dstStandardHour
                                        // dstStandardMinute
                                        // dstStandardSecond
										// dstDaylightMonth
										// dstDaylightDate
										// dstDaylightYear
										// dstDaylightDayOfWeek
                                        // dstDaylightHour
                                        // dstDaylightMinute
                                        // dstDaylightSecond
                                        //
										// Response: "Y" (ie. ACK)
#define PC_GET_CONFIG_REQUEST       'n' // Get the misc. Configuration settings
#define PC_GET_CONFIG_RESPONSE      'o' //
                                        // request: "n"
                                        // response: "o<see PC_SET_CONFIG request>"
                                        // where:
                                        //       see PC_SET_CONFIG request
                                        //
#define PC_SET_NMEA_CONFIG          'p' // Set the PCIe port to use for NMEA messages from the GPS receiver
                                        // request: "p<port,baud,OnOff,message,message,...>"
                                        // where:
                                        // port == 0,PCIe port 0 (primary serial port, also used for communication with DLL)
                                        //                        If primary is selected, ZDA is internally generated and
                                        //                        does not come from the GPS receiver.
                                        //         1,PCIe port 1 (secondary serial port, only used for NMEA messages,
                                        //                        connected directly to GPS receiver)
                                        // The baud rates below are only used if PCIe port 1 (secondary) is specified.
                                        // baud == 4800(default),9600,19200,38400,57600,115200
                                        //         The baud rates above are only used if PCIe port 1 (secondary) is specified.
                                        //         baud rate is ignored if port == 1
                                        //         8 data bits, no parity, 1 stop bit cannot be changed.
                                        // OnOff == 0, turn off all NMEA messages
                                        //             if 0 is specified, "message" is ignored.
                                        // OnOff == 1, turn on selected messages
                                        // message == 1,ZDA
                                        //            2,GGA
                                        //            3,GLL
                                        //            4,VTG
                                        //            5,GSA
                                        //            6,GSV
                                        //            7,GST
                                        //            8,RMC
										// Response: "Y" (ie. ACK)
#define PC_GET_NMEA_CONFIG_REQUEST  'q' // Get the settings for the PCIe port to use for NMEA messages from the GPS receiver
#define PC_GET_NMEA_CONFIG_RESPONSE 'r' //
                                        // request: "q"
                                        // response: "r<port,baud,OnOff,message,message,...>"
                                        // where:
                                        // port    == see PC_SET_NMEA_CONFIG
                                        // baud    == see PC_SET_NMEA_CONFIG
                                        // OnOff   == see PC_SET_NMEA_CONFIG
                                        // message == see PC_SET_NMEA_CONFIG
#define PC_SET_GPS_FIX_CONFIG       's' // Set the GPS navigation fix settings.
                                        // request: "s<sensitivity,dynamics,elevation,SBAS>"
                                        // see Navigation TSIP 0xBB
                                        // where:
                                        // sensitivity == 0, standard(Copernicus II default) TSIP 0x69
                                        //             == 1, (Masterclock default)
                                        //  Dynamics code: 1,land(default) vehicle type of environment,
                                        //                 2,sea-low acceleration environment,
                                        //                 3,air-high acceleration environment,
                                        //                 4,stationary
                                        //  elevation mask: 5 degree, .0873 radians(Copernicus II default)
                                        //                  2 degree, .0349 radians(Masterclock default)
                                        //                  Note:: The AVR receives the decimal equivalent value and passes
                                        //                         it to the Copernicus II unchanged. In other words, we define this as an
                                        //                         ascii string equivalent to the %d output of the single precision
                                        //                         floating point number, save as an UINT32 and just pass the
                                        //                         32 bit value.
                                        //                         On the PC,
                                        //                         UINT32 myHexRadians;
                                        //                         float myRadians;
                                        //                         char cAsciiString[64];
                                        //                         myRadians = .0349;
                                        //                         myHexRadians = (UINT32)myRadians;
                                        //                         sprintf_s(cAsciiString,sizeof(cAsciiString),"%lu",myHexRadians);
                                        //                         // cAsciiString is "elevation" sent to the AVR.
                                        //  SBAS on/off: 0,WAAS_OFF
                                        //               1,WAAS_auto (default)
										// Response: "Y" (ie. ACK)
#define PC_GET_GPS_FIX_CONFIG_REQUEST  't' // Get the GPS navigation fix settings.
#define PC_GET_GPS_FIX_CONFIG_RESPONSE 'u' //
                                        // request: "t"
                                        // response: "u<sensitivity,dynamics,elevation>"
                                        // where:
                                        // sensitivity == see PC_SET_GPS_FIX_CONFIG
                                        // dynamics    == see PC_SET_GPS_FIX_CONFIG
                                        // elevation   == see PC_SET_GPS_FIX_CONFIG
#define PC_GPS_SATELLITES_REQUEST   'v' // Request signal levels for all satellites being tracked
#define PC_GPS_SATELLITES_RESPONSE  'w' // Equivalent to TRIMBLE_PKT_C_SIGLEV
                                        // Request: "v"
										// Response: "w<>" (data received from TRIMBLE_PKT_R_SIGLEV)
                                        // Where
#define PC_ACK						'Y'	// (ie. ACK) acknowledge of receiving and processing a valid command
#define PC_NAK						'Z'	// (ie. NAK) last command rejected
										// "Z<command,reason>"
										// where "commmand" is the single character command ID,
										// "reason" is a numeric value indicating one of the following:
										// 1 ==command not recognized
										// 2 ==bad parameter
										// 3 ==too many parameters
										// 4 ==not enough parameters


// bit mapped flags
#define REF_IS_LOCKED       0x80
#define REF_IS_UNLOCKED     0x00
#define REF_LOCKED_MASK     0x80
#define NUM_SATELLITES_MASK 0x0F
#define TIME_CODE_TYPE_MASK 0x0F

// time code type
#define CALIBRATING         0x00
#define SMPTE_30            0x01
#define SMPTE_25            0x02
#define SMPTE_24            0x03
#define IRIG_B_PWM          0x04
#define IRIG_B_AM           0x05

// note: These defines must match defines in IntApiDefs.h API_LOCK_INVALID, API_LOCK_NOT_PRESENT, API_LOCK_LOST, API_LOCK_RESTORED, API_LOCK_PRESENT
#define LOCK_INVALID         0x00   // never had a lock
#define LOCK_NOT_PRESENT     0x01   // had a lock but lost it
#define LOCK_LOST            0x02   // notify, one time, that lock was lost
#define LOCK_RESTORED        0x03   // notify, one time, that lock was restored
#define LOCK_PRESENT         0x04   // currently locked to a reference

#define VALID_ALARM_ENTRY	0x09 // kpw - WARNING - This is added to replace the above as a validity check in fnProcessAVR_NOTIFY_ALARM

// note: These defines must match the defines in IntApiDefs.h: API_HW_STATUS_NORMAL, API_HW_STATUS_CARDCOMMFAIL 
#define HW_STATUS_NORMAL       0
#define HW_STATUS_FAILED       1

typedef struct _LAST_LOCK_STATUS
{
	UINT8 ucLockStatus;
	UINT8 ucRtcStatus;
	UINT8 ucGpsStatus;
	UINT8 ucTimeCodeStatus;
	UINT8 ucHsoStatus;
    UINT8 ucHsoAchievementLevel;
	UINT8 ucReserved1;
	UINT8 ucReserved2;
	UINT8 ucReserved3;
	UINT8 ucReserved4;
} LAST_LOCK_STATUS;

typedef struct _LAST_ALARM_STATUS
{
	LAST_LOCK_STATUS srLockStatus;
	UINT8 ucHwStatus;
	UINT8 verMajor;
	UINT8 verMinor;
	UINT8 verRevision;
	UINT8 ucReserved1;
	UINT8 ucReserved2;
	UINT8 ucReserved3;
	UINT8 ucReserved4;
} LAST_ALARM_STATUS;


typedef struct _TIME_CODE_FRAME_COMPLETE
{
	LARGE_INTEGER   intPerfCount;   // performance counter before frame was read

	LAST_LOCK_STATUS   srLockStatus;   // The overall lock status, includes time code, GPS, HSO, and RTC

	UINT8			tcStatus;		// time/time code status when the frame was read

	UINT8           hwStatus;		// hardware status when the frame was read

	UINT8			tcFrame[13];	// Buffer for timecode to be read into, large enough for
	                                // a complete SMPTE or IRIG frame

} TIME_CODE_FRAME_COMPLETE, *P_TIME_CODE_FRAME_COMPLETE;



// These are used as bits to indicate an active NMEA packet
//#define TIME_NMEA_BIT   0x0001
#define GGA_NMEA_BIT	0x0002
#define GLL_NMEA_BIT	0x0004
#define GSA_NMEA_BIT	0x0008
#define GSV_1_NMEA_BIT	0x0010
#define GSV_2_NMEA_BIT	0x0020
#define GSV_3_NMEA_BIT	0x0040
#define RMC_NMEA_BIT	0x0080
#define VTG_NMEA_BIT	0x0100
#define ZDA_NMEA_BIT	0x0200



#define GPS_PRIMARY_PORT_A   0
#define GPS_SECONDARY_PORT_B 1



extern LAST_ALARM_STATUS g_LastAlarmStatusResponse;

INT   McrSerialInterruptDSR_GetTime(LARGE_INTEGER* pliPerfCountIn);
BOOL  fnSerialComPortInit(CSerial *pCMcrSerialPort, int iPortNum);
void  fnSerialEventMonitorThread(void);
void  fnSerialTopOfSecondThread(void);
void  fnSerialTimeCodeFrameAlarmThread(void);
BOOL  fnReadTimeCodeFrame(TIME_CODE_FRAME_COMPLETE *psrPreviousSecondTimeFrame);
void  fnSetAllAvrConfiguration(void);
BOOL  fnRecalibrate(void);
BOOL  fnSetDateTime(MCRSYNC_IPC_SETTIME *psrSetTime);
void  CloseSerialEvents(void);
BOOL  fnValidateComPort(CMcrSerialInterface *pCMcrSerialInterface, CSoftwareConfig *pswConfig);
BOOL  fnIsComPortValid(CMcrSerialInterface *pCMcrSerialInterface, int iComPortNumber);
BOOL  fnIsNmeaComPortValid(CMcrSerialInterface *pCMcrSerialInterface, int iComPortNumber);
BOOL  fnOpenComPort(CSerial *pCSerial, int iComPortNumber, int iBaudRate, int iDataBits, int iParity, int iStopBits);
void  fnConfigureNmeaMessages(CSerial *pCSerial, int iBaudRate, DWORD dwNmeaMessagesBitMap, int iIsVirtualComPort);
//INT   EnableMcrInterrupt(DWORD intToEnable);
//INT   DisableMcrInterrupt(DWORD intToEnable);